import base64
import graphene


def _resolve_type_coordinates(self, info):
    return self['coordinates']


def _resolve_fs_field(field, name, default_value=None):
    v = getattr(field.instance, field.key)
    return getattr(v, name, default_value)


def _resolve_content_type(self, info):
    return _resolve_fs_field(self, 'content_type')


def _resolve_md5(self, info):
    return _resolve_fs_field(self, 'md5')


def _resolve_chunk_size(self, info):
    return _resolve_fs_field(self, 'chunk_size', 0)


def _resolve_length(self, info):
    return _resolve_fs_field(self, 'length', 0)


def _resolve_data(self, info):
    v = getattr(self.instance, self.key)
    data = v.read()
    if data is not None:
        return base64.b64encode(data)
    return None


class FileFieldType(graphene.ObjectType):

    content_type = graphene.String(resolver=_resolve_content_type)
    md5 = graphene.String(resolver=_resolve_md5)
    chunk_size = graphene.Int(resolver=_resolve_chunk_size)
    length = graphene.Int(resolver=_resolve_length)
    data = graphene.String(resolver=_resolve_data)


class _TypeField(graphene.ObjectType):

    type = graphene.String()

    def resolve_type(self, info):
        return self['type']


class PointFieldType(_TypeField):

    coordinates = graphene.List(
        graphene.Float, resolver=_resolve_type_coordinates)


class PolygonFieldType(_TypeField):

    coordinates = graphene.List(
        graphene.List(
            graphene.List(graphene.Float)),
        resolver=_resolve_type_coordinates
    )


class MultiPolygonFieldType(_TypeField):

    coordinates = graphene.List(
        graphene.List(
            graphene.List(
                graphene.List(graphene.Float))),
        resolver=_resolve_type_coordinates)



import mongoengine
import uuid
from graphene import (
    ID,
    Boolean,
    DateTime,
    Dynamic,
    Field,
    Float,
    Int,
    List,
    NonNull,
    String,
    Union,
    is_node
)
from graphene.types.json import JSONString
from mongoengine.base import get_document

from . import advanced_types
from .utils import (
    import_single_dispatch, get_field_description,
)

singledispatch = import_single_dispatch()


class MongoEngineConversionError(Exception):
    pass


@singledispatch
def convert_mongoengine_field(field, registry=None):
    raise MongoEngineConversionError(
        "Don't know how to convert the MongoEngine field %s (%s)" %
        (field, field.__class__))


@convert_mongoengine_field.register(mongoengine.EmailField)
@convert_mongoengine_field.register(mongoengine.StringField)
@convert_mongoengine_field.register(mongoengine.URLField)
def convert_field_to_string(field, registry=None):
    return String(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.UUIDField)
@convert_mongoengine_field.register(mongoengine.ObjectIdField)
def convert_field_to_id(field, registry=None):
    return ID(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.IntField)
@convert_mongoengine_field.register(mongoengine.LongField)
def convert_field_to_int(field, registry=None):
    return Int(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.BooleanField)
def convert_field_to_boolean(field, registry=None):
    return Boolean(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.DecimalField)
@convert_mongoengine_field.register(mongoengine.FloatField)
def convert_field_to_float(field, registry=None):
    return Float(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.DateTimeField)
def convert_field_to_datetime(field, registry=None):
    return DateTime(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.DictField)
@convert_mongoengine_field.register(mongoengine.MapField)
def convert_field_to_jsonstring(field, registry=None):
    return JSONString(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.PointField)
def convert_point_to_field(field, registry=None):
    return Field(advanced_types.PointFieldType)


@convert_mongoengine_field.register(mongoengine.PolygonField)
def convert_polygon_to_field(field, registry=None):
    return Field(advanced_types.PolygonFieldType)


@convert_mongoengine_field.register(mongoengine.MultiPolygonField)
def convert_multipolygon_to_field(field, register=None):
    return Field(advanced_types.MultiPolygonFieldType)


@convert_mongoengine_field.register(mongoengine.FileField)
def convert_file_to_field(field, registry=None):
    return Field(advanced_types.FileFieldType)


@convert_mongoengine_field.register(mongoengine.ListField)
@convert_mongoengine_field.register(mongoengine.EmbeddedDocumentListField)
def convert_field_to_list(field, registry=None):
    base_type = convert_mongoengine_field(field.field, registry=registry)
    if isinstance(base_type, (Dynamic)):
        base_type = base_type.get_type()
        if base_type is None:
            return
        base_type = base_type._type

    if is_node(base_type):
        return base_type._meta.connection_field_class(base_type)

    # Non-relationship field
    relations = (mongoengine.ReferenceField, mongoengine.EmbeddedDocumentField)
    if not isinstance(base_type, (List, NonNull)) \
            and not isinstance(field.field, relations):
        base_type = type(base_type)

    return List(base_type, description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.GenericReferenceField)
def convert_field_to_union(field, registry=None):

    _types = []
    for choice in field.choices:
        _field = mongoengine.ReferenceField(get_document(choice))
        _field = convert_mongoengine_field(_field, registry)
        _type = _field.get_type()
        if _type:
            _types.append(_type.type)
        else:
            # TODO: Register type auto-matically here.
            pass

    if len(_types) == 0:
        return None

    # XXX: Use uuid to avoid duplicate name
    name = '{}_{}_union_{}'.format(
        field._owner_document.__name__,
        field.db_field,
        str(uuid.uuid1()).replace('-', '')
    )
    Meta = type('Meta', (object, ), {'types': tuple(_types)})
    _union = type(name, (Union, ), {'Meta': Meta})
    return Field(_union)


@convert_mongoengine_field.register(mongoengine.EmbeddedDocumentField)
@convert_mongoengine_field.register(mongoengine.ReferenceField)
def convert_field_to_dynamic(field, registry=None):
    model = field.document_type

    def dynamic_type():
        _type = registry.get_type_for_model(model)
        if not _type:
            return None
        return Field(_type, description=get_field_description(field, registry))

    return Dynamic(dynamic_type)


@convert_mongoengine_field.register(mongoengine.LazyReferenceField)
def convert_lazy_field_to_dynamic(field, registry=None):
    model = field.document_type

    def lazy_resolver(root, *args, **kwargs):
        if getattr(root, field.name or field.db_name):
            return getattr(root, field.name or field.db_name).fetch()

    def dynamic_type():
        _type = registry.get_type_for_model(model)
        if not _type:
            return None
        return Field(_type, resolver=lazy_resolver, description=get_field_description(field, registry))

    return Dynamic(dynamic_type)



from __future__ import absolute_import

from collections import OrderedDict
from functools import partial, reduce

import graphene
import mongoengine
from graphene.relay import ConnectionField
from graphene.types.argument import to_arguments
from graphene.types.dynamic import Dynamic
from graphene.types.structures import Structure
from graphql_relay.connection.arrayconnection import connection_from_list_slice

from .advanced_types import (
    FileFieldType, PointFieldType, MultiPolygonFieldType
)
from .converter import convert_mongoengine_field, MongoEngineConversionError
from .registry import get_global_registry
from .utils import get_model_reference_fields, get_node_from_global_id


class MongoengineConnectionField(ConnectionField):

    def __init__(self, type, *args, **kwargs):
        get_queryset = kwargs.pop('get_queryset', None)
        if get_queryset:
            assert callable(get_queryset), "Attribute `get_queryset` on {} must be callable.".format(self)
        self._get_queryset = get_queryset
        super(MongoengineConnectionField, self).__init__(
            type,
            *args,
            **kwargs
        )

    @property
    def type(self):
        from .types import MongoengineObjectType
        _type = super(ConnectionField, self).type
        assert issubclass(
            _type, MongoengineObjectType), "MongoengineConnectionField only accepts MongoengineObjectType types"
        assert _type._meta.connection, "The type {} doesn't have a connection".format(
            _type.__name__)
        return _type._meta.connection

    @property
    def node_type(self):
        return self.type._meta.node

    @property
    def model(self):
        return self.node_type._meta.model

    @property
    def registry(self):
        return getattr(self.node_type._meta, 'registry', get_global_registry())

    @property
    def args(self):
        return to_arguments(
            self._base_args or OrderedDict(),
            dict(self.field_args, **self.reference_args)
        )

    @args.setter
    def args(self, args):
        self._base_args = args

    def _field_args(self, items):

        def is_filterable(k):
            """
            Args:
                k (str): field name.
            Returns:
                bool
            """

            if not hasattr(self.model, k):
                return False
            if isinstance(getattr(self.model, k), property):
                return False
            try:
                converted = convert_mongoengine_field(getattr(self.model, k), self.registry)
            except MongoEngineConversionError:
                return False
            if isinstance(converted, (ConnectionField, Dynamic)):
                return False
            if callable(getattr(converted, 'type', None)) \
                    and isinstance(
                        converted.type(),
                        (FileFieldType, PointFieldType, MultiPolygonFieldType, graphene.Union)):
                return False
            return True

        def get_type(v):
            if isinstance(v.type, Structure):
                return v.type.of_type()
            return v.type()

        return {k: get_type(v) for k, v in items if is_filterable(k)}

    @property
    def field_args(self):
        return self._field_args(self.fields.items())

    @property
    def reference_args(self):

        def get_reference_field(r, kv):
            field = kv[1]
            mongo_field = getattr(self.model, kv[0], None)
            if isinstance(mongo_field, (mongoengine.LazyReferenceField, mongoengine.ReferenceField)):
                field = convert_mongoengine_field(mongo_field, self.registry)
            if callable(getattr(field, 'get_type', None)):
                _type = field.get_type()
                if _type:
                    node = _type._type._meta
                    if 'id' in node.fields and not issubclass(node.model, (mongoengine.EmbeddedDocument,)):
                        r.update({kv[0]: node.fields['id']._type.of_type()})
            return r

        return reduce(get_reference_field, self.fields.items(), {})

    @property
    def fields(self):
        return self._type._meta.fields

    def get_queryset(self, model, info, **args):

        if args:
            reference_fields = get_model_reference_fields(self.model)
            hydrated_references = {}
            for arg_name, arg in args.copy().items():
                if arg_name in reference_fields:
                    reference_obj = get_node_from_global_id(reference_fields[arg_name], info, args.pop(arg_name))
                    hydrated_references[arg_name] = reference_obj
            args.update(hydrated_references)
        if self._get_queryset:
            queryset_or_filters = self._get_queryset(model, info, **args)
            if isinstance(queryset_or_filters, mongoengine.QuerySet):
                return queryset_or_filters
            else:
                args.update(queryset_or_filters)
        return model.objects(**args)

    def default_resolver(self, _root, info, **args):
        args = args or {}

        connection_args = {
            'first': args.pop('first', None),
            'last': args.pop('last', None),
            'before': args.pop('before', None),
            'after': args.pop('after', None)
        }

        _id = args.pop('id', None)

        if _id is not None:
            objs = [get_node_from_global_id(self.node_type, info, _id)]
            list_length = 1
        elif callable(getattr(self.model, 'objects', None)):
            objs = self.get_queryset(self.model, info, **args)
            list_length = objs.count()
        else:
            objs = []
            list_length = 0

        connection = connection_from_list_slice(
            list_slice=objs,
            args=connection_args,
            list_length=list_length,
            connection_type=self.type,
            edge_type=self.type.Edge,
            pageinfo_type=graphene.PageInfo,
        )
        connection.iterable = objs
        connection.list_length = list_length
        return connection

    def chained_resolver(self, resolver, root, info, **args):
        resolved = resolver(root, info, **args)
        if resolved is not None:
            return resolved
        return self.default_resolver(root, info, **args)

    def get_resolver(self, parent_resolver):
        super_resolver = self.resolver or parent_resolver
        resolver = partial(self.chained_resolver, super_resolver)
        return partial(self.connection_resolver, resolver, self.type)
